/* R. Harmsen, October-November 2023
   First completely working version 15 December 2023.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "rhmusicadhoc.h"

struct IntvalRatio PlausibIntvals_00[] = {
   { 1, 1}, {-1,-1}};
struct IntvalRatio PlausibIntvals_01[] = {
   {16,15}, {17,16}, {18,17}, {19,18}, {20,19}, {25,24}, {-1,-1}};
struct IntvalRatio PlausibIntvals_02[] = {
   { 9, 8}, {10, 9}, {19,17}, {-1,-1}};
struct IntvalRatio PlausibIntvals_03[] = {
   { 6, 5}, {19,16}, {20,17}, {13,11}, {32,27}, {25,21}, {-1,-1}};
struct IntvalRatio PlausibIntvals_04[] = {
   { 5, 4}, {24,19}, {81,64}, {19,15}, {14,11}, {-1,-1}};
struct IntvalRatio PlausibIntvals_05[] = {
   { 4, 3}, {-1,-1}};
struct IntvalRatio PlausibIntvals_06[] = {
   {24,17}, {17,12}, {45,32}, {64,45}, { 7, 5}, {10, 7}, {-1,-1}};
struct IntvalRatio PlausibIntvals_07[] = {
   { 3, 2}, {-1,-1}};
struct IntvalRatio PlausibIntvals_08[] = {
   { 8, 5}, {19,12}, {27,17}, {11, 7}, {128,81}, {-1,-1}};
struct IntvalRatio PlausibIntvals_09[] = {
   { 5, 3}, {27,16}, {32,19}, {22,13}, {17,10}, {-1,-1}};
struct IntvalRatio PlausibIntvals_10[] = {
   { 9, 5}, {16, 9}, {-1,-1}};
struct IntvalRatio PlausibIntvals_11[] = {
   {15, 8}, {17, 9}, {19,10}, {21,11}, {48,25}, {-1,-1}};
struct IntvalRatio PlausibIntvals_12[] = {
   { 2, 1}, {-1,-1}};
struct IntvalRatio PlausibIntvals_13[] = {
   {2*16,15}, {17,16/2}, {2*18,17}, {19,18/2}, {2*20,19}, {25,24/2}, {-1,-1}};
struct IntvalRatio PlausibIntvals_14[] = {
   { 9, 8/2}, {2*10, 9}, {2*19,17}, {-1,-1}};

struct IntvalRatio PlausibIntvals_00_5[] = {
   {28,27}, {27,26}, {-1,-1}};
struct IntvalRatio PlausibIntvals_01_5[] = {
   {12,11}, {11,10}, {-1,-1}};
struct IntvalRatio PlausibIntvals_02_5[] = {
   {15,13}, {22,19}, { 8, 7}, { 7, 6}, {-1,-1}};
struct IntvalRatio PlausibIntvals_03_5[] = {
   {11, 9}, {27,22}, {16,13}, {21,17}, {-1,-1}};
struct IntvalRatio PlausibIntvals_04_5[] = {
   {22,17}, {13,10}, {17,13}, { 9, 7}, {-1,-1}};
struct IntvalRatio PlausibIntvals_05_5[] = {
   {26,19}, {15,11}, {11, 8}, {18,13}, {25,18}, {-1,-1}};
struct IntvalRatio PlausibIntvals_06_5[] = {
   {13, 9}, {16,11}, {19,13}, {22,15}, {25,17}, {-1,-1}};
struct IntvalRatio PlausibIntvals_07_5[] = {
   {20,13}, {17,11}, {14, 9}, {-1,-1}};
struct IntvalRatio PlausibIntvals_08_5[] = {
   {13, 8}, {18,11}, {33,20}, {28,17}, {-1,-1}};
struct IntvalRatio PlausibIntvals_09_5[] = {
   {12, 7}, {19,11}, {26,15}, { 7, 4}, {-1,-1}};
struct IntvalRatio PlausibIntvals_10_5[] = {
   {20,11}, {11, 6}, {24,13}, {13, 7}, {-1,-1}};
struct IntvalRatio PlausibIntvals_11_5[] = {
   {25,13}, {27,14}, {-1,-1}};
struct IntvalRatio PlausibIntvals_12_5[] = {
   {2*28,27}, {27,26/2}, {-1,-1}};
struct IntvalRatio PlausibIntvals_13_5[] = {
   {2*12,11}, {11,10/2}, {-1,-1}};
struct IntvalRatio PlausibIntvals_14_5[] = {
   {2*15,13}, {2*22,19}, {2*8, 7}, { 7, 6/2}, {-1,-1}};

struct IntvalRatio *ChordIntv[NumIntvals] =
{
   PlausibIntvals_00, /* Prime */
   PlausibIntvals_00_5,
   PlausibIntvals_01, /* Minor second */
   PlausibIntvals_01_5,
   PlausibIntvals_02, /* Major second */
   PlausibIntvals_02_5,
   PlausibIntvals_03, /* Minor third */
   PlausibIntvals_03_5,
   PlausibIntvals_04,
   PlausibIntvals_04_5,
   PlausibIntvals_05, /* Fourth */
   PlausibIntvals_05_5,
   PlausibIntvals_06,
   PlausibIntvals_06_5,
   PlausibIntvals_07, /* Fifth */
   PlausibIntvals_07_5,
   PlausibIntvals_08,
   PlausibIntvals_08_5,
   PlausibIntvals_09, /* Major sixth */
   PlausibIntvals_09_5,
   PlausibIntvals_10, /* Minor seventh */
   PlausibIntvals_10_5,
   PlausibIntvals_11,
   PlausibIntvals_11_5,
   PlausibIntvals_12, /* Octave */
   PlausibIntvals_12_5,
   PlausibIntvals_13,
   PlausibIntvals_13_5,
   PlausibIntvals_14, /* Major ninth */
   PlausibIntvals_14_5,
};

static long attempts = 0;

static long int Optim = 100000;

static void AdhocRecurs (int *Chord, int NotesInChord,
   struct DegreeData *wrkC, struct DegreeData *resC, int grado);
static int  FindRatioSequence (struct DegreeData *wrkC, int maxgrad);
static void AdhocReport (int *Chord, int NotesInChord,
   struct DegreeData *resC, int verbose);
static void AdhocDoReport (
   int quartertones, struct DegreeData *repC, int verbose);

/***********************************************************************/
void AdhocRunIt (int *Chord, int NotesInChord,
   struct DegreeData *wrkC, struct DegreeData *resC, int verbose)
{
   int i;

   attempts = 0;
   Optim = 100000;

   for (i = 0; i < NotesInChord; i++)
      wrkC[i].Degree =
         Chord[i] < NumIntvals ? ChordIntv[Chord[i]] : NULL;
   AdhocRecurs(Chord, NotesInChord, wrkC, resC, 0);
   AdhocReport(Chord, NotesInChord,       resC, verbose);
}

/***********************************************************************/
static void AdhocRecurs (int *Chord, int NotesInChord,
   struct DegreeData *wrkC, struct DegreeData *resC, int grado)
{
   int ivl;

   if (grado >= NotesInChord || wrkC[grado].Degree == NULL)
      return;

   for (ivl = 0; wrkC[grado].Degree[ivl].numer > 0; ivl++)
   {
      wrkC[grado].Considered.numer = wrkC[grado].Degree[ivl].numer;
      wrkC[grado].Considered.denom = wrkC[grado].Degree[ivl].denom;

      if (grado == NotesInChord - 1 || wrkC[grado + 1].Degree == NULL)
      {
         attempts++;
         FindRatioSequence(wrkC, grado);
         #ifdef DEBUG
         for (int i = 0; i <= grado; i++)
            fprintf(stdout, "grado=%d, RatSeq %5lld %3ld:%-3ld %4.0f\n", i,
               wrkC[i].RatioSequence,
               wrkC[i].Considered.numer,
               wrkC[i].Considered.denom,
               rats2cents(wrkC[i].Considered.numer, wrkC[i].Considered.denom)) ;
         fprintf(stdout, "\n");
         #endif

         if (wrkC[0].RatioSequence < Optim)
         {
            memcpy(resC, wrkC, sizeof *resC * NotesInChord);
            Optim = wrkC[0].RatioSequence;
         }
      }
      AdhocRecurs(Chord, NotesInChord, wrkC, resC, grado + 1);
   }
}

/***********************************************************************/
static int FindRatioSequence (struct DegreeData *wrkC, int maxgrad)
{
   int grado;
   /* Least common multiple */
   LLI lcm = wrkC[0].Considered.denom;
   LLI numerGcd = wrkC[0].Considered.numer;
   LLI thisGcd;

   for (grado = 1; grado <= maxgrad; grado++)
   {
      numerGcd = gcd(wrkC[grado].Considered.numer, numerGcd);
      thisGcd  = gcd(wrkC[grado].Considered.denom, lcm);
      #ifdef DEBUG
      fprintf(stdout, "lcm=%4lld, considering %ld, gcd=%lld\n",
         lcm, wrkC[grado].Considered.denom, thisGcd);
      #endif
      lcm = lcm * wrkC[grado].Considered.denom / thisGcd;
      #ifdef DEBUG
      fprintf(stdout, "new lcm=%4lld\n\n", lcm);
      #endif
   }

   for (grado = 0; grado <= maxgrad; grado++)
   {
      #ifdef DEBUG
      fprintf(stdout, "RatSeq %3lld = numerator %3ld * lcm %3lld / denom %3ld\n",
         wrkC[grado].Considered.numer * lcm / wrkC[grado].Considered.denom,
         wrkC[grado].Considered.numer , lcm , wrkC[grado].Considered.denom);
      #endif
      wrkC[grado].RatioSequence =
      wrkC[grado].Considered.numer * lcm / wrkC[grado].Considered.denom / numerGcd;
   }
   return 0;
}

/***********************************************************************/
static void AdhocReport (int *Chord, int NotesInChord,
   struct DegreeData *resC, int verbose)
{
   int i;
   struct DegreeData *repC =
      calloc(NumIntvals, sizeof (struct DegreeData));

   if (!repC) return;

   if (verbose > 3)
   {
      fprintf(stdout, "Attempts: %ld\nResult\n", attempts);
      for (i = 0; i < NotesInChord && resC[i].Degree; i++)
      {
         fprintf(stdout, "grado=%d, RatSeq %5lld %3ld:%-3ld %4.0f\n", i,
            resC[i].RatioSequence,
            resC[i].Considered.numer,
            resC[i].Considered.denom,
            rats2cents(resC[i].Considered.numer, resC[i].Considered.denom));
      }
   }

   for (i = 0; i < NotesInChord; i++)
   {
      if (Chord[i] < NumIntvals)
         repC[Chord[i]] = resC[i];
   }
   AdhocDoReport(0, repC, verbose);
   AdhocDoReport(1, repC, verbose);

   free(repC);
}

/***********************************************************************/
static void AdhocDoReport (
   int quartertones, struct DegreeData *repC, int verbose)
{
   int i, quartertone_present = 0;

   if (quartertones)
   {
      for (i = 1; i < NumIntvals; i += 2)
         if (repC[i].RatioSequence)
         {
            quartertone_present = 1;
            break;
         }
      if (!quartertone_present)
         return;
   }

   if (verbose > 1)
   {
      if (quartertones)
         printf("    ");
      for (i = quartertones ? 1 : 0; i < NumIntvals; i += 2)
      {
         if (repC[i].RatioSequence)
            printf("%4lld    ", repC[i].RatioSequence);
         else
            printf("        ");
      }
      printf("\n");
   }

   if (verbose > 0)
   {
      if (quartertones)
         printf("    ");
      for (i = quartertones ? 1 : 0; i < NumIntvals; i += 2)
      {
         if (repC[i].RatioSequence)
            printf("%3ld:%-3ld ", repC[i].Considered.numer,
                                  repC[i].Considered.denom);
         else
            printf("        ");
      }
      printf("\n");
   }

   if (verbose > 1)
   {
      if (quartertones)
         printf("    ");
      for (i = quartertones ? 1 : 0; i < NumIntvals; i += 2)
      {
         if (repC[i].RatioSequence)
            printf(" %4.0f   ",
               rats2cents(repC[i].Considered.numer, repC[i].Considered.denom));
         else
            printf("        ");
      }
      printf("\n\n");
   }
}

/***********************************************************************/
#ifdef WOLF_FIFTHS_LIST

#define FIFTHRATIO (3.0 / 2.0)
#define FORTHRATIO (4.0 / 3.0)
#define DIDYMEAN_COMMA (81.0 / 80.0)
#define UPPER5 (FIFTHRATIO / DIDYMEAN_COMMA) * 1.0001
#define LOWER5 (FIFTHRATIO / DIDYMEAN_COMMA) / 1.0001
#define UPPER4 (FORTHRATIO * DIDYMEAN_COMMA) * 1.0001
#define LOWER4 (FORTHRATIO * DIDYMEAN_COMMA) / 1.0001

int main (void)
{
   int grado, other, o, i, j;

   for (grado = 0; grado < NumIntvals; grado++)
   {
      other = grado + 7*2;
      for (o = other - 1; o <= other + 1; o++)
      {
         int oo = o % (12*2);
         double r1, r2, tp;

         fprintf(stdout, "\nComparing intervals of %4.1f and %4.1f\n",
            grado / 2.0, oo / 2.0);
         for (i = 0; ChordIntv[grado][i].numer > 0; i++)
         for (j = 0; ChordIntv[oo]   [j].numer > 0; j++)
         {
            fprintf(stdout, "Intervals %3ld:%-3ld %3ld:%-3ld\n",
              ChordIntv[oo]   [j].numer,
              ChordIntv[oo]   [j].denom,
              ChordIntv[grado][i].numer,
              ChordIntv[grado][i].denom);
            r1 = ChordIntv[oo]   [j].numer / (double)ChordIntv[oo]   [j].denom;
            r2 = ChordIntv[grado][i].numer / (double)ChordIntv[grado][i].denom;
            if (r2 > r1)
            {
               tp = r1;
               r1 = r2;
               r2 = tp;
            }
            fprintf(stdout, "Found ratios %6.3f / %6.3f = %6.3f\n", r1, r2, r1 / r2);
            if (r1 / r2 > LOWER4 && r1 / r2 < UPPER4)
               fprintf(stdout, "Found WOLF FOURTH!! %8.3f cents\n",
                  rats2cents (r1, r2));
            if (r1 / r2 > LOWER5 && r1 / r2 < UPPER5)
               fprintf(stdout, "Found WOLF FIFTH!!! %8.3f cents\n",
                  rats2cents (r1, r2));
         }
      }
   }
   return 0;
}
#endif
